/*
 * Name: OGRegularExpressionMatch.h
 * Project: OgreKit
 *
 * Creation Date: Aug 30 2003
 * Author: Isao Sonobe <sonoisa (AT) muse (DOT) ocn (DOT) ne (DOT) jp>
 * Copyright: Copyright (c) 2003 Isao Sonobe, All rights reserved.
 * License: OgreKit License
 *
 * Encoding: UTF8
 * Tabsize: 4
 */

#import <Foundation/Foundation.h>
#ifndef NOT_RUBY
#	define NOT_RUBY
#endif
#ifndef HAVE_CONFIG_H
#	define HAVE_CONFIG_H
#endif
#import <OgreKit/oniguruma.h>


// constant
extern NSString	* const OgreMatchException;


@class OGRegularExpression, OGRegularExpressionEnumerator, OGRegularExpressionCapture;
@protocol OGStringProtocol;

@interface OGRegularExpressionMatch : NSObject <NSCopying, NSCoding>
{
	OnigRegion		*_region;						// match result region
	OGRegularExpressionEnumerator	*_enumerator;	// matcher
	unsigned		_terminalOfLastMatch;           // Position to the end of the last matched string (_region->end[0] / sizeof(unichar))
	
	NSObject<OGStringProtocol>	*_targetString;		// Search string
	NSRange			_searchRange;					// Search scope
	unsigned		_index;							// Matched order
}

/*********
 * Various Information *
 *********/
// Matched order 0,1,2,...
- (unsigned)index;

// The number of substrings + 1
- (unsigned)count;

// description
- (NSString*)description;


/*********
 * String *
 *********/
// Been the subject of a string match
- (NSObject<OGStringProtocol>*)targetOGString;
- (NSString*)targetString;
- (NSAttributedString*)targetAttributedString;

// Match string \&, \0
- (NSObject<OGStringProtocol>*)matchedOGString;
- (NSString*)matchedString;
- (NSAttributedString*)matchedAttributedString;

// The second substring index \index
//  The second index when there is no substring returns nil.
- (NSObject<OGStringProtocol>*)ogSubstringAtIndex:(unsigned)index;
- (NSString*)substringAtIndex:(unsigned)index;
- (NSAttributedString*)attributedSubstringAtIndex:(unsigned)index;

// String before the matched substring \ `
- (NSObject<OGStringProtocol>*)prematchOGString;
- (NSString*)prematchString;
- (NSAttributedString*)prematchAttributedString;

// After the matched substring from a string \'
- (NSObject<OGStringProtocol>*)postmatchOGString;
- (NSString*)postmatchString;
- (NSAttributedString*)postmatchAttributedString;

// 最後にマッチした部分文字列 \+
// 存在しないときには nil を返す。
- (NSObject<OGStringProtocol>*)lastMatchOGSubstring;
- (NSString*)lastMatchSubstring;
- (NSAttributedString*)lastMatchAttributedSubstring;

// マッチした部分と一つ前にマッチした部分の間の文字列 \- (独自に追加)
- (NSObject<OGStringProtocol>*)ogStringBetweenMatchAndLastMatch;
- (NSString*)stringBetweenMatchAndLastMatch;
- (NSAttributedString*)attributedStringBetweenMatchAndLastMatch;


/*******
 * 範囲 *
 *******/
// Range of the matched string
- (NSRange)rangeOfMatchedString;

// substring index th range
// The second substring index does not exist at (-1, 0) is returned.
- (NSRange)rangeOfSubstringAtIndex:(unsigned)index;

// Range of the string before the matched substring
- (NSRange)rangeOfPrematchString;

// マッチした部分より後ろの文字列の範囲
- (NSRange)rangeOfPostmatchString;

// 最後にマッチした部分文字列の範囲
// 存在しないときには {-1,0} を返す。
- (NSRange)rangeOfLastMatchSubstring;

// マッチした部分と一つ前にマッチした部分の間の文字列の範囲
- (NSRange)rangeOfStringBetweenMatchAndLastMatch;


/***************************************************************
 * Related named group (OgreCaptureGroupOption available when specified) *
 ***************************************************************/
// The name (label) is part of the name string
// If the name does not exist, returns nil.
// If there are multiple substrings with the same name, raises an exception.
- (NSObject<OGStringProtocol>*)ogSubstringNamed:(NSString*)name;
- (NSString*)substringNamed:(NSString*)name;
- (NSAttributedString*)attributedSubstringNamed:(NSString*)name;

// The scope of the name name substring
// If the name does not exist (-1, 0) is returned.
// If there are multiple substrings with the same name, raises an exception.
- (NSRange)rangeOfSubstringNamed:(NSString*)name;

// The index of the substring with the supplied name
// If the name does not exist returns -1.
// If there are multiple substrings with the same name, raises an exception.
- (unsigned)indexOfSubstringNamed:(NSString*)name;

// The name of the index-th substring
// If the name does not exist, returns nil.
- (NSString*)nameOfSubstringAtIndex:(unsigned)index;

/***********************
* Get matched substring *
************************/
// (Regex1) | (regex2 )|... such as regular expressions, regex * useful for any conditional or matched by
/* Example: 
	OGRegularExpression *regex = [OGRegularExpression regularExpressionWithString:@"([0-9]+)|([a-zA-Z]+)"];
	NSEnumerator	*matchEnum = [regex matchEnumeratorInString:@"123abc"];
	OGRegularExpressionMatch	*match;
	while ((match = [matchEnum nextObject]) != nil) {
		switch ([match indexOfFirstMatchedSubstring]) {
			case 1:
				NSLog(@"numbers");
				break;
			case 2:
				NSLog(@"alphabets");
				break;
		}
	}
*/
// First position of substring matched by group (if not return 0)
- (unsigned)indexOfFirstMatchedSubstring;
- (unsigned)indexOfFirstMatchedSubstringBeforeIndex:(unsigned)anIndex;
- (unsigned)indexOfFirstMatchedSubstringAfterIndex:(unsigned)anIndex;
- (unsigned)indexOfFirstMatchedSubstringInRange:(NSRange)aRange;
// Name
- (NSString*)nameOfFirstMatchedSubstring;
- (NSString*)nameOfFirstMatchedSubstringBeforeIndex:(unsigned)anIndex;
- (NSString*)nameOfFirstMatchedSubstringAfterIndex:(unsigned)anIndex;
- (NSString*)nameOfFirstMatchedSubstringInRange:(NSRange)aRange;

// マッチした部分文字列のうちグループ番号が最大のもの (ない場合は0を返す)
- (unsigned)indexOfLastMatchedSubstring;
- (unsigned)indexOfLastMatchedSubstringBeforeIndex:(unsigned)anIndex;
- (unsigned)indexOfLastMatchedSubstringAfterIndex:(unsigned)anIndex;
- (unsigned)indexOfLastMatchedSubstringInRange:(NSRange)aRange;
// その名前
- (NSString*)nameOfLastMatchedSubstring;
- (NSString*)nameOfLastMatchedSubstringBeforeIndex:(unsigned)anIndex;
- (NSString*)nameOfLastMatchedSubstringAfterIndex:(unsigned)anIndex;
- (NSString*)nameOfLastMatchedSubstringInRange:(NSRange)aRange;

// マッチした部分文字列のうち最長のもの (ない場合は0を返す。同じ長さの物が複数あれば、番号の小さい物が優先される)
- (unsigned)indexOfLongestSubstring;
- (unsigned)indexOfLongestSubstringBeforeIndex:(unsigned)anIndex;
- (unsigned)indexOfLongestSubstringAfterIndex:(unsigned)anIndex;
- (unsigned)indexOfLongestSubstringInRange:(NSRange)aRange;
// その名前
- (NSString*)nameOfLongestSubstring;
- (NSString*)nameOfLongestSubstringBeforeIndex:(unsigned)anIndex;
- (NSString*)nameOfLongestSubstringAfterIndex:(unsigned)anIndex;
- (NSString*)nameOfLongestSubstringInRange:(NSRange)aRange;

// マッチした部分文字列のうち最短のもの (ない場合は0を返す。同じ長さの物が複数あれば、番号の小さい物が優先される)
- (unsigned)indexOfShortestSubstring;
- (unsigned)indexOfShortestSubstringBeforeIndex:(unsigned)anIndex;
- (unsigned)indexOfShortestSubstringAfterIndex:(unsigned)anIndex;
- (unsigned)indexOfShortestSubstringInRange:(NSRange)aRange;
// その名前
- (NSString*)nameOfShortestSubstring;
- (NSString*)nameOfShortestSubstringBeforeIndex:(unsigned)anIndex;
- (NSString*)nameOfShortestSubstringAfterIndex:(unsigned)anIndex;
- (NSString*)nameOfShortestSubstringInRange:(NSRange)aRange;

/******************
* Capture History *
*******************/
/*例:
	NSString					*target = @"abc de";
	OGRegularExpression			*regex = [OGRegularExpression regularExpressionWithString:@"(?@[a-z])+"];
	OGRegularExpressionMatch	*match;
    OGRegularExpressionCapture  *capture;
	NSEnumerator				*matchEnumerator = [regex matchEnumeratorInString:target];
	unsigned					i;
	
	while ((match = [matchEnumerator nextObject]) != nil) {
		capture = [match captureHistory];
		NSLog(@"number of capture history: %d", [capture numberOfChildren]);
		for (i = 0; i < [capture numberOfChildren]; i++) 
            NSLog(@" %@", [[capture childAtIndex:i] string]);
	}
	
ログ:
number of capture history: 3
 a
 b
 c
number of capture history: 2
 d
 e
 */

// 捕獲履歴
// 履歴がない場合はnilを返す。
- (OGRegularExpressionCapture*)captureHistory;

@end

// UTF16文字列の長さを得る
inline unsigned Ogre_UTF16strlen(unichar *const aUTF16string, unichar *const end);
